"""
Example file showing how to perform a finite-difference (FD) step size study
using WISDEM and then visualize the results.
"""


import os

import numpy as np

from wisdem import run_wisdem

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = os.path.join(os.path.dirname(mydir), "05_tower_monopile", "nrel5mw_tower.yaml")
fname_modeling_options = mydir + os.sep + "modeling_options.yaml"
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, analysis_options, opt_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)


# Load the derivative information
all_derivs = np.load("total_derivs.npy", allow_pickle=True)[()]

# Select a key of interest; an (of, wrt) pair
key = ("towerse.post.constr_stress", "tower.diameter")

# Collect data to plot later
step_sizes = []
deriv_values = []

for idx in all_derivs:
    derivs = all_derivs[idx]
    step_sizes.append(derivs["step_size"])
    deriv_values.append(derivs[key][0][0])

print("Step size | Deriv")
for step_size, deriv in zip(step_sizes, deriv_values):
    print(f"{step_size:06}     {deriv}")

try:
    # Plot the step size study results
    import matplotlib.pyplot as plt

    plt.plot(step_sizes, deriv_values)
    plt.xlabel("FD step size")
    plt.ylabel("Derivative value of stress wrt diameter")
    plt.gca().set_xscale("log")

    plt.gca().invert_xaxis()

    plt.show()
except Exception:
    pass



import unittest

import numpy as np
import numpy.testing as npt

from wisdem.commonse.cross_sections import Tube, Rectangle

npts = 100


class TestTube(unittest.TestCase):
    def testTubeScalar(self):
        T = Tube(2 * 5.0, 1.0)

        self.assertAlmostEqual(T.Area, np.pi * 9.0)
        self.assertAlmostEqual(T.Ixx, np.pi * 369.0 / 4.0)
        self.assertAlmostEqual(T.Iyy, np.pi * 369.0 / 4.0)
        self.assertAlmostEqual(T.J0, np.pi * 369.0 / 2.0)
        self.assertAlmostEqual(T.S, np.pi * 369.0 / 4.0 / 5.0)
        self.assertAlmostEqual(T.C, np.pi * 369.0 / 2.0 / 5.0)

    def testTubeVector(self):
        n = 5
        my1 = np.ones(n)
        T = Tube(2 * 5.0 * my1, my1)

        npt.assert_almost_equal(T.Area, np.pi * 9.0)
        npt.assert_almost_equal(T.Ixx, np.pi * 369.0 / 4.0)
        npt.assert_almost_equal(T.Iyy, np.pi * 369.0 / 4.0)
        npt.assert_almost_equal(T.J0, np.pi * 369.0 / 2.0)
        npt.assert_almost_equal(T.S, np.pi * 369.0 / 4.0 / 5.0)
        npt.assert_almost_equal(T.C, np.pi * 369.0 / 2.0 / 5.0)

class TestRectangle(unittest.TestCase):
    def testRectangleScalar(self):
        R = Rectangle(2.0, 4.0, 0.5)

        self.assertAlmostEqual(R.Area, 5.0)
        self.assertAlmostEqual(R.Ixx, 8.41666667)
        self.assertAlmostEqual(R.Iyy, 2.41666667)
        self.assertAlmostEqual(R.J0, 10.83333333)
        self.assertAlmostEqual(R.BdgMxx, 4.20833333)
        self.assertAlmostEqual(R.BdgMyy, 2.41666667)

    def testRectangleError(self):
        try:
            Rectangle(1.0, 2.0, 3.0)
        except Exception as err:
            assert isinstance(err, AssertionError)
        else:
            raise AssertionError("1, 2, 3, should not define a rectangular propertly")

    def testRectangleVector(self):
        n = 5
        my1 = np.ones(n)
        R = Rectangle(2.0 * my1, 4.0 * my1, 0.5 * my1)

        npt.assert_almost_equal(R.Area, 5.0)
        npt.assert_almost_equal(R.Ixx, 8.41666667)
        npt.assert_almost_equal(R.Iyy, 2.41666667)
        npt.assert_almost_equal(R.J0, 10.83333333)
        npt.assert_almost_equal(R.BdgMxx, 4.20833333)
        npt.assert_almost_equal(R.BdgMyy, 2.41666667)




if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from wisdem.commonse.distribution import WeibullCDF, RayleighCDF, WeibullWithMeanCDF

np.random.seed(314)


class Test(unittest.TestCase):
    def test_distributions(self):
        nspline = 10

        prob = om.Problem(reports=False)

        prob.model.add_subsystem("comp1", WeibullCDF(nspline=nspline), promotes_inputs=["*"])
        prob.model.add_subsystem("comp2", WeibullWithMeanCDF(nspline=nspline), promotes_inputs=["*"])
        prob.model.add_subsystem("comp3", RayleighCDF(nspline=nspline), promotes_inputs=["*"])

        prob.setup()

        # Add some arbitrary inputs
        prob.set_val("x", np.random.rand(nspline), units="m/s")
        prob.set_val("xbar", 1.5, units="m/s")
        prob.set_val("k", 1.5)
        prob.set_val("A", 1.2)

        prob.run_model()

        check = prob.check_partials(out_stream=None, compact_print=True, method="fd", step=1e-4)

        assert_check_partials(check, atol=1e-3, rtol=1e-4)


if __name__ == "__main__":
    unittest.main()



"""Provides a testing framework for the `Cable` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


import copy
import itertools

import numpy as np
import pytest

from wisdem.orbit.phases.design._cables import Cable, Plant

cables = {
    "empty": {},
    "passes": {
        "conductor_size": 400,
        "current_capacity": 610,
        "rated_voltage": 33,
        "ac_resistance": 0.06,
        "inductance": 0.375,
        "capacitance": 225,
        "linear_density": 35,
        "cost_per_km": 300000,
        "name": "passes",
    },
}

plant_config_calculate_all_ring = {
    "site": {"depth": 20},
    "plant": {
        "layout": "ring",
        "row_spacing": 7,
        "turbine_spacing": 5,
        "num_turbines": 40,
    },
    "turbine": {"rotor_diameter": 154, "turbine_rating": 10},
}

plant_config_calculate_all_grid = {
    "site": {"depth": 20},
    "plant": {
        "layout": "grid",
        "row_spacing": 7,
        "turbine_spacing": 5,
        "num_turbines": 40,
    },
    "turbine": {"rotor_diameter": 154, "turbine_rating": 10},
}

plant_config_distance_provided_ring = {
    "site": {"depth": 20},
    "plant": {
        "layout": "ring",
        "row_distance": 0.4,
        "turbine_distance": 0.1,
        "num_turbines": 40,
        "substation_distance": 0.2,
    },
    "turbine": {"rotor_diameter": 154, "turbine_rating": 10},
}

plant_config_distance_provided_grid = {
    "site": {"depth": 20},
    "plant": {
        "layout": "grid",
        "row_distance": 0.4,
        "turbine_distance": 0.1,
        "num_turbines": 40,
        "substation_distance": 0.2,
    },
    "turbine": {"rotor_diameter": 154, "turbine_rating": 10},
}

plant_config_custom = {
    "site": {"depth": 20},
    "plant": {
        "layout": "custom",
        "row_distance": 0.4,
        "turbine_distance": 0.1,
        "num_turbines": 40,
        "substation_distance": 0.2,
    },
    "turbine": {"rotor_diameter": 154, "turbine_rating": 10},
}


def test_cable_creation():
    cable = Cable(cables["passes"])

    assert cable
    for r in cable.required:
        assert getattr(cable, r, None) == cables["passes"][r]


def test_cable_required_inputs():
    with pytest.raises(ValueError):
        Cable(cables["empty"])


def test_power_factor():
    c = copy.deepcopy(cables["passes"])

    results = []
    for i in itertools.product(
        range(100, 1001, 150),  # conductor size
        np.arange(0.01, 0.91, 0.1),  # ac_resistance
        np.arange(0, 1, 0.15),  # inductance
        range(100, 1001, 150),  # capacitance
    ):
        c["conductor_size"] = i[0]
        c["ac_resistance"] = i[1]
        c["inductance"] = i[2]
        c["capacitance"] = i[3]

        cable = Cable(c)
        results.append(cable.power_factor)

    if any((a < 0) | (a > 1) for a in results):
        raise Exception("Invalid Power Factor.")


@pytest.mark.parametrize(
    "config",
    (
        plant_config_calculate_all_ring,
        plant_config_calculate_all_grid,
        plant_config_distance_provided_ring,
        plant_config_distance_provided_grid,
    ),
    ids=["calculate_ring", "calculate_grid", "provided_ring", "provided_grid"],
)
def test_plant_creation(config):
    plant = Plant(config)

    assert plant.turbine_rating == config["turbine"]["turbine_rating"]
    assert plant.site_depth == config["site"]["depth"] / 1000.0
    assert plant.layout == config["plant"]["layout"]
    assert plant.num_turbines == config["plant"]["num_turbines"]

    if "turbine_spacing" in config["plant"]:
        td = config["turbine"]["rotor_diameter"] * config["plant"]["turbine_spacing"] / 1000.0
    else:
        td = config["plant"]["turbine_distance"]
    assert plant.turbine_distance == td

    if "row_spacing" in config["plant"]:
        if config["plant"]["layout"] == "grid":
            rd = config["turbine"]["rotor_diameter"] * config["plant"]["row_spacing"] / 1000.0
        if config["plant"]["layout"] == "ring":
            rd = td
    else:
        rd = config["plant"]["row_distance"]
    assert plant.row_distance == rd

    if "substation_distance" in config["plant"]:
        sd = config["plant"]["substation_distance"]
    else:
        sd = td
    assert plant.substation_distance == sd


def test_custom_plant_creation():
    plant = Plant(plant_config_custom)

    for attr in ("row_distance", "turbine_distance", "substation_distance"):
        assert getattr(plant, attr, None) is None



"""Test suite for the library module."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"

import os
from copy import deepcopy

import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.core import library
from wisdem.orbit.core.exceptions import LibraryItemNotFoundError

config = {
    "site": {"distance": 100, "depth": 15},
    "plant": {"num_turbines": 10},
    "turbine": {
        "hub_height": 100,
        "tower": {"type": "Tower", "deck_space": 100, "mass": 400},
        "nacelle": {"type": "Nacelle", "deck_space": 200, "mass": 400},
        "blade": {"type": "Blade", "deck_space": 100, "mass": 100},
    },
    "MonopileInstallation": {"wtiv": "test_wtiv"},
    "TurbineInstallation": {"wtiv": "test_wtiv"},
    "port": {"num_cranes": 1, "monthly_rate": 10000},
    "monopile": {
        "type": "Monopile",
        "length": 50,
        "diameter": 10,
        "deck_space": 500,
        "mass": 350,
    },
    "transition_piece": {
        "type": "Transition Piece",
        "deck_space": 250,
        "mass": 350,
    },
    "install_phases": ["MonopileInstallation", "TurbineInstallation"],
}


def test_initialize_library():
    library.initialize_library(None)
    assert os.environ["DATA_LIBRARY"]

    _ = os.environ.pop("DATA_LIBRARY")
    library.initialize_library(pytest.library)
    assert os.environ["DATA_LIBRARY"] == pytest.library


def test_extract_library_specs_fail():
    library.initialize_library(pytest.library)
    with pytest.raises(LibraryItemNotFoundError):
        library.extract_library_specs("turbine", "unknown.yaml")


def test_phase_specific_file_extraction():
    project = ProjectManager(config)
    turbine_config = project.create_config_for_phase("TurbineInstallation")
    monopile_config = project.create_config_for_phase("MonopileInstallation")

    assert isinstance(turbine_config["wtiv"], dict)
    assert isinstance(monopile_config["wtiv"], dict)

    bad_config = deepcopy(config)
    _ = bad_config.pop("TurbineInstallation")
    bad_config["wtiv"] = "example_wtiv"
    bad_config["MonopileInstallation"]["wtiv"] = "missing_vessel"

    with pytest.raises(LibraryItemNotFoundError):
        bad_project = ProjectManager(bad_config)



import numpy as np
import openmdao.api as om

import wisdem.commonse.fileIO as fio
from wisdem.drivetrainse.generator import Generator

opt_flag = False
n_pc = 20

# Example optimization of a generator for costs on a 5 MW reference turbine
prob = om.Problem(reports=False)
prob.model = Generator(design="dfig", n_pc=n_pc)

if opt_flag:
    # add optimizer and set-up problem (using user defined input on objective function)
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"

    # Specificiency target efficiency(%)
    Eta_Target = 0.930

    eps = 1e-6

    # Design variables
    prob.model.add_design_var("r_s", lower=0.2, upper=1.0)
    prob.model.add_design_var("len_s", lower=0.4, upper=2.0)
    prob.model.add_design_var("h_s", lower=0.04, upper=0.1)
    prob.model.add_design_var("B_symax", lower=1.0, upper=2.0 - eps)
    prob.model.add_design_var("I_0", lower=5.0, upper=100.0, ref=1e2)
    prob.model.add_design_var("S_Nmax", lower=-0.3, upper=-0.1)

    # Constraints
    prob.model.add_constraint("generator_efficiency", lower=Eta_Target, indices=[-1])
    prob.model.add_constraint("E_p", lower=500.0 + eps, upper=5000.0 - eps, ref=1e3)
    prob.model.add_constraint("TCr", lower=0.0 + eps)
    prob.model.add_constraint("TCs", lower=0.0 + eps)
    prob.model.add_constraint("B_g", lower=0.7, upper=1.2)
    prob.model.add_constraint("B_trmax", upper=2.0 - eps)
    prob.model.add_constraint("B_tsmax", upper=2.0 - eps)
    prob.model.add_constraint("A_1", upper=60000.0 - eps, ref=1e5, indices=[-1])
    prob.model.add_constraint("J_s", upper=6.0)
    prob.model.add_constraint("J_r", upper=6.0)
    prob.model.add_constraint("Slot_aspect_ratio1", lower=4.0, upper=10.0)
    prob.model.add_constraint("K_rad", lower=0.2, upper=1.5)
    prob.model.add_constraint("D_ratio", lower=1.37, upper=1.4)
    prob.model.add_constraint("Current_ratio", lower=0.1, upper=0.3)

    Objective_function = "generator_cost"
    prob.model.add_objective(Objective_function, scaler=1e-5)


prob.setup()

# Specify Target machine parameters

prob["machine_rating"] = 5000000.0
prob["shaft_rpm"] = np.linspace(200, 1200.0, n_pc)
prob["cofi"] = 0.9
prob["y_tau_p"] = 12.0 / 15.0
prob["sigma"] = 21.5e3
prob["rad_ag"] = 0.61  # 0.493167295965 #0.61 #meter
prob["len_s"] = 0.49  # 1.06173588215 #0.49 #meter
prob["h_s"] = 0.08  # 0.1 # 0.08 #meter
prob["I_0"] = 40.0  # 40.0191207049 #40.0 #Ampere
prob["B_symax"] = 1.3  # 1.59611292026 #1.3 #Tesla
prob["S_Nmax"] = -0.2  # -0.3 #-0.2
prob["k_fillr"] = 0.55
prob["q1"] = 5
prob["h_0"] = 0.1
# ---------------------------------------------------
prob["E"] = 2e11
prob["G"] = 79.3e9

# Specific costs
prob["C_Cu"] = 4.786  # Unit cost of Copper $/kg
prob["C_Fe"] = 0.556  # Unit cost of Iron $/kg
prob["C_Fes"] = 0.50139  # specific cost of Structural_mass $/kg
prob["C_PM"] = 95.0

# Material properties
prob["rho_Fe"] = 7700.0  # Steel density Kg/m3
prob["rho_Fes"] = 7850  # structural Steel density Kg/m3
prob["rho_Copper"] = 8900.0  # copper density Kg/m3
prob["rho_PM"] = 7450.0  # typical density Kg/m3 of neodymium magnets (added 2019 09 18) - for pmsg_[disc|arms]

# Run optimization
if opt_flag:
    prob.model.approx_totals()
    prob.run_driver()
else:
    prob.run_model()

prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)
# fio.save_data('DFIG', prob, npz_file=False, mat_file=False, xls_file=True)
print("Efficiency table:")
print(" rpm     converter  transformer  generator   ")
print("-------------------------------------------")
print(np.c_[prob["shaft_rpm"],
            prob["converter_efficiency"],
            prob["transformer_efficiency"],
            prob["generator_efficiency"],
            ])



"""Provides the `Port` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import simpy

from wisdem.orbit.core.exceptions import ItemNotFound


class Port(simpy.FilterStore):
    """Port Class"""

    def __init__(self, env, **kwargs):
        """
        Creates an instance of Port.

        Parameters
        ----------
        env : simpy.Environment
            SimPy environment that simulation runs on.
        """

        capacity = kwargs.get("capacity", float("inf"))
        super().__init__(env, capacity)

    def get_item(self, _type):
        """
        Checks self.items for an item satisfying `item.type = _type`, otherwise
        returns `ItemNotFound`.

        Parameters
        ----------
        _type : str
            Type of item to match. Checks `item.type`.

        Returns
        -------
        res.value : FilterStoreGet.value
            Returned item.

        Raises
        ------
        ItemNotFound
        """

        target = None
        for i in self.items:
            try:
                if i.type == _type:
                    target = i
                    break

            except AttributeError:
                continue

        if not target:
            raise ItemNotFound(_type)

        else:
            res = self.get(lambda x: x == target)
            return res.value


class WetStorage(simpy.Store):
    """Storage infrastructure for floating substructures."""

    def __init__(self, env, capacity):
        """
        Creates an instance of WetStorage.

        Parameters
        ----------
        capacity : int
            Number of substructures or assemblies that can be stored.
        """

        super().__init__(env, capacity)



# 1 ---------

from math import pi

import numpy as np
import matplotlib.pyplot as plt

from wisdem.ccblade.ccblade import CCBlade, CCAirfoil

# geometry
Rhub = 1.5
Rtip = 63.0

r = np.array(
    [
        2.8667,
        5.6000,
        8.3333,
        11.7500,
        15.8500,
        19.9500,
        24.0500,
        28.1500,
        32.2500,
        36.3500,
        40.4500,
        44.5500,
        48.6500,
        52.7500,
        56.1667,
        58.9000,
        61.6333,
    ]
)
chord = np.array(
    [
        3.542,
        3.854,
        4.167,
        4.557,
        4.652,
        4.458,
        4.249,
        4.007,
        3.748,
        3.502,
        3.256,
        3.010,
        2.764,
        2.518,
        2.313,
        2.086,
        1.419,
    ]
)
theta = np.array(
    [
        13.308,
        13.308,
        13.308,
        13.308,
        11.480,
        10.162,
        9.011,
        7.795,
        6.544,
        5.361,
        4.188,
        3.125,
        2.319,
        1.526,
        0.863,
        0.370,
        0.106,
    ]
)
B = 3  # number of blades

tilt = 5.0
precone = 2.5
yaw = 0.0

nSector = 8  # azimuthal discretization

# atmosphere
rho = 1.225
mu = 1.81206e-5

# power-law wind shear profile
shearExp = 0.2
hubHt = 90.0

# 1 ----------


# 2 ----------
import os
from wisdem.inputs.validation import load_geometry_yaml
baseyaml = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "02_reference_turbines", "nrel5mw.yaml")
data = load_geometry_yaml(baseyaml)
af = data['airfoils']
af_names = ["Cylinder", "Cylinder", "DU40_A17", "DU35_A17", "DU30_A17", "DU25_A17", "DU21_A17", "NACA64_A17"]
airfoil_types = [0] * len(af_names)
for i in range(len(af_names)):
    for j in range(len(af)):
        if af[j]["name"] == af_names[i]:
            polars = af[j]['polars'][0]
            airfoil_types[i] = CCAirfoil(
                np.rad2deg(polars["c_l"]["grid"]),
                [polars["re"]],
                polars["c_l"]["values"],
                polars["c_d"]["values"],
                polars["c_m"]["values"],
            )

# place at appropriate radial stations
af_idx = [0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7]

af = [0] * len(r)
for i in range(len(r)):
    af[i] = airfoil_types[af_idx[i]]

# 2 ----------

# 3 ----------

# create CCBlade object
rotor = CCBlade(
    r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp, hubHt, nSector, derivatives=True
)

# 3 ----------

# 4 ----------
# set conditions
Uinf = 10.0
tsr = 7.55
pitch = 0.0
Omega = Uinf * tsr / Rtip * 30.0 / pi  # convert to RPM
azimuth = 0.0

# # evaluate distributed loads
# Np, Tp = rotor.distributedAeroLoads(Uinf, Omega, pitch, azimuth)

# 4 ----------

# 5 ----------

loads, derivs = rotor.distributedAeroLoads(Uinf, Omega, pitch, azimuth)
Np = loads["Np"]
Tp = loads["Tp"]
dNp = derivs["dNp"]
dTp = derivs["dTp"]

n = len(r)

# n x n (diagonal)
dNp_dr = dNp["dr"]
dNp_dchord = dNp["dchord"]
dNp_dtheta = dNp["dtheta"]
dNp_dpresweep = dNp["dpresweep"]

# n x n (tridiagonal)
dNp_dprecurve = dNp["dprecurve"]

# n x 1
dNp_dRhub = dNp["dRhub"]
dNp_dRtip = dNp["dRtip"]
dNp_dprecone = dNp["dprecone"]
dNp_dtilt = dNp["dtilt"]
dNp_dhubHt = dNp["dhubHt"]
dNp_dyaw = dNp["dyaw"]
dNp_dazimuth = dNp["dazimuth"]
dNp_dUinf = dNp["dUinf"]
dNp_dOmega = dNp["dOmega"]
dNp_dpitch = dNp["dpitch"]

# 5 ----------


# 6 ----------


loads, derivs = rotor.evaluate([Uinf], [Omega], [pitch])
P = loads["P"]
T = loads["T"]
Q = loads["Q"]

dP = derivs["dP"]
dT = derivs["dT"]
dQ = derivs["dQ"]

npts = len(P)

# npts x 1
dP_dprecone = dP["dprecone"]
dP_dtilt = dP["dtilt"]
dP_dhubHt = dP["dhubHt"]
dP_dRhub = dP["dRhub"]
dP_dRtip = dP["dRtip"]
dP_dprecurveTip = dP["dprecurveTip"]
dP_dpresweepTip = dP["dpresweepTip"]
dP_dyaw = dP["dyaw"]

# npts x npts
dP_dUinf = dP["dUinf"]
dP_dOmega = dP["dOmega"]
dP_dpitch = dP["dpitch"]

# npts x n
dP_dr = dP["dr"]
dP_dchord = dP["dchord"]
dP_dtheta = dP["dtheta"]
dP_dprecurve = dP["dprecurve"]
dP_dpresweep = dP["dpresweep"]


# 6 ----------


# 7 ----------

idx = 8
delta = 1e-6 * r[idx]
r[idx] += delta

rotor_fd = CCBlade(
    r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp, hubHt, nSector, derivatives=False
)

r[idx] -= delta

loads, derivs = rotor_fd.distributedAeroLoads(Uinf, Omega, pitch, azimuth)
Npd = loads["Np"]
Tpd = loads["Tp"]

dNp_dr_fd = (Npd - Np) / delta
dTp_dr_fd = (Tpd - Tp) / delta

print("(analytic) dNp_i/dr_i =", dNp_dr[idx, idx])
print("(fin diff) dNp_i/dr_i =", dNp_dr_fd[idx])
print()
# 7 ----------


# 8 ----------
delta = 1e-6 * precone
precone += delta

rotor_fd = CCBlade(
    r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp, hubHt, nSector, derivatives=False
)

precone -= delta

loads, derivs = rotor_fd.evaluate([Uinf], [Omega], [pitch], coefficients=False)
Pd = loads["P"]
Td = loads["T"]
Qd = loads["Q"]

dT_dprecone_fd = (Td - T) / delta
dQ_dprecone_fd = (Qd - Q) / delta
dP_dprecone_fd = (Pd - P) / delta

print("(analytic) dP/dprecone =", dP_dprecone[0, 0])
print("(fin diff) dP/dprecone =", dP_dprecone_fd[0])
print()
# 8 ----------


# 9 ----------
idx = 12
delta = 1e-6 * r[idx]
r[idx] += delta

rotor_fd = CCBlade(
    r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp, hubHt, nSector, derivatives=False
)

r[idx] -= delta

loads, derivs = rotor_fd.evaluate([Uinf], [Omega], [pitch], coefficients=False)
Pd = loads["P"]
Td = loads["T"]
Qd = loads["Q"]

dT_dr_fd = (Td - T) / delta
dQ_dr_fd = (Qd - Q) / delta
dP_dr_fd = (Pd - P) / delta

print("(analytic) dP/dr_i =", dP_dr[0, idx])
print("(fin diff) dP/dr_i =", dP_dr_fd[0])
print()
# 9 ----------



__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = ["jake.nunemaker@nrel.gov"]


import re
import time
from copy import deepcopy
from random import sample
from itertools import product

import yaml
import numpy as np
import pandas as pd
import statsmodels.api as sm
from yaml import Loader
from benedict import benedict

from wisdem.orbit import ProjectManager


class ParametricManager:
    """Class for configuring parametric ORBIT runs."""

    def __init__(
        self,
        base,
        params,
        funcs,
        weather=None,
        module=None,
        product=False,
        keep_inputs=[],
    ):
        """
        Creates an instance of `ParametricRun`.

        Parameters
        ----------
        base : dict
            Base ORBIT configuration.
        params : dict
            Parameters and their values.
            Format: "subdict.param": [num1, num2, num3]
        funcs : dict
            List of functions used to save results.
            Format: "name": lambda project: project.attr
        weather : DataFrame

        """

        self.base = benedict(base)
        self.params = params
        self.funcs = funcs
        self.weather = weather
        self.results = None
        self.module = module
        self.product = product
        self.keep = keep_inputs

    def run(self, **kwargs):
        """Run the configured parametric runs and save any requested results to
        `self.results`."""

        outputs = []
        for run in self.run_list:
            data = self._run_config(run, **kwargs)
            outputs.append(data)

        self.results = pd.DataFrame(outputs)

    def _run_config(self, run, **kwargs):
        """Run an individual config."""

        config = deepcopy(self.base)
        config.merge(run)

        if self.module is not None:
            project = self.module(config, weather=self.weather, **kwargs)
            project.run()

        else:
            project = ProjectManager(config, weather=self.weather, **kwargs)
            project.run()

        results = self.map_funcs(project, self.funcs)
        kept = self._get_kept_inputs(project.config)
        data = {**run, **kept, **results}

        return data

    def _get_kept_inputs(self, config):
        """
        Extract inputs in `self.keep` from `config`.

        Parameters
        ----------
        config : dict
        """

        kept = {}
        for k in self.keep:
            try:
                kept[k] = config[k]

            except KeyError:
                pass

        return kept

    @property
    def run_list(self):
        """Returns list of configured parametric runs."""

        if self.product:
            runs = list(product(*self.params.values()))

        else:
            runs = list(zip(*self.params.values()))

        return [dict(zip(self.params.keys(), run)) for run in runs]

    @property
    def num_runs(self):
        return len(self.run_list)

    @staticmethod
    def map_funcs(obj, funcs):
        """
        Map `obj` to list of `funcs`.

        Parameters
        ----------
        obj : ProjectManager
            Project instance to run through functions.
        funcs : list
            Functions used to pull results from obj.
        """

        results = {}
        for k, f in funcs.items():
            try:
                res = f(obj)

            except TypeError:
                raise TypeError(
                    f"Result function '{f}' not structured properly. " f"Correct format: 'lambda project: project.{f}'"
                )

            except AttributeError:
                res = np.NaN

            results[k] = res

        return results

    def preview(self, num=10, **kwargs):
        """
        Runs a limited set of runs to preview the results and provide an
        estimate for total run time.

        Parameters
        ----------
        num : int
            Number to run.
        """

        start = time.time()
        outputs = []
        if num > len(self.run_list):
            to_run = self.run_list

        else:
            to_run = sample(self.run_list, num)

        for run in to_run:
            data = self._run_config(run, **kwargs)
            outputs.append(data)

        elapsed = time.time() - start
        estimate = (len(self.run_list) / num) * elapsed
        print(f"{num} runs elapsed time: {elapsed:.2f}s")
        print(f"{self.num_runs} runs estimated time: {estimate:.2f}s")

        return pd.DataFrame(outputs)

    def create_model(self, x, y):
        """"""

        if self.results is None:
            print("`ParametricManager hasn't been ran yet.")

        return LinearModel(self.results, x, y)

    @classmethod
    def from_config(cls, data):
        """"""

        outputs = data.pop("outputs", {})

        funcs = {}
        for k, v in outputs.items():
            split = v.split("[")
            attr = split[0]

            try:
                key = re.sub("[^A-Za-z ]", "", split[1])

            except IndexError:
                key = None

            if key is None:
                funcs[k] = lambda run, a=attr: getattr(run, a)

            else:
                funcs[k] = lambda run, a=attr, k=key: getattr(run, a)[k]

        data["funcs"] = funcs

        module_name = data.pop("module", None)
        if module_name is not None:
            module = ProjectManager.phase_dict()[module_name]
            data["module"] = module

        weather_file = data.pop("weather", None)
        if weather_file is not None:
            weather = pd.read_csv(weather_file, parse_dates=["datetime"])
            data["weather"] = weather

        return cls(**data, product=True)


class LinearModel:
    """Simple linear regression model."""

    def __init__(self, data, x, y):
        """
        Creates an instance of `LinearModel`.

        Parameters
        ----------
        data : pd.DataFrame
            Input dataset.
        x : list
            List of regression variables.
        y : str
            Output variable.
        """

        self.data = data
        self.x = x
        self.y = y

        self.X = data[x]
        self.Y = data[y]

        self.X2 = sm.add_constant(self.X)
        self.sm = sm.OLS(self.Y, self.X2).fit()

    def predict(self, inputs):
        """
        Predicts the output value of `inputs` given the underlying linear model
        developed using `self.data`, `self.x` and `self.y`.

        Parameters
        ----------
        inputs : dict
            Inputs with form:
            'key': list | np.array | int | float
        """

        missing = [i for i in self.x if i not in inputs.keys()]
        if len(missing) > 0:
            raise ValueError(f"Missing input(s) '{missing}'")

        inputs = {k: np.array(v) for k, v in inputs.items()}
        params = deepcopy(self.sm.params)
        const = params.pop("const")

        return sum([v * inputs[k] for k, v in params.items()]) + const

    @property
    def as_string(self):
        """
        Returns the linear model represented as a string. To be used for
        interfacing with NRWAL.

        Returns
        -------
        str
        """

        params = deepcopy(self.sm.params)
        const = params.pop("const")

        out = ""
        for i, (k, v) in enumerate(params.items()):
            if i == 0:
                pre = ""

            else:
                if v >= 0:
                    pre = " + "

                else:
                    pre = " - "

            out += pre
            out += f"{abs(v)} * {k}"

        if const < 0:
            out += f" - {abs(const)}"

        else:
            out += f" + {const}"

        return out

    @property
    def vif(self):
        """
        Returns the variance inflation factor of the input data series.

        Returns
        -------
        list
        """

        data = self.X2.copy()

        vif = []
        for name, d in data.iteritems():
            r_sq_i = sm.OLS(d, data.drop(name, axis=1)).fit().rsquared
            vif.append(1.0 / (1.0 - r_sq_i))

        return vif

    @property
    def perc_diff(self):
        """
        Returns the percent difference between the predicted values and the
        output value.

        Returns
        -------
        pd.Series
        """

        inputs = dict(zip(self.X.T.index, self.X.T.values))
        predicted = self.predict(inputs)

        return (self.Y - predicted) / self.Y



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import pandas as pd
import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.phases import DesignPhase, InstallPhase
from wisdem.orbit.manager import ProjectProgress
from wisdem.orbit.core.library import extract_library_specs
from wisdem.test.test_orbit.data import test_weather
from wisdem.orbit.core.exceptions import MissingInputs, PhaseNotFound, WeatherProfileError, PhaseDependenciesInvalid

weather_df = pd.DataFrame(test_weather).set_index("datetime")

config = extract_library_specs("config", "project_manager")
complete_project = extract_library_specs("config", "complete_project")


### Top Level
@pytest.mark.parametrize("weather", (None, weather_df))
def test_complete_run(weather):
    project = ProjectManager(config, weather=weather)
    project.run()

    actions = pd.DataFrame(project.actions)

    phases = ["MonopileInstallation", "TurbineInstallation"]
    assert all(p in list(actions["phase"]) for p in phases)


### Module Integrations
def test_for_required_phase_structure():
    """
    Automated integration test to verify that all classes listed in
    ProjectManager.possible_phases are structured correctly.
    """

    for p in ProjectManager._install_phases:
        assert isinstance(p.expected_config, dict)

    for p in ProjectManager._design_phases:
        assert isinstance(p.expected_config, dict)
        assert isinstance(p.output_config, dict)


# TODO: Expand these tests


### Config Management
def test_phase_specific_definitions():
    """
    Tests that phase specific information makes it to phase_config.
    """

    project = ProjectManager(config)

    phase_config = project.create_config_for_phase("MonopileInstallation")

    assert phase_config["wtiv"]["name"] == "Phase Specific WTIV"
    assert phase_config["site"]["distance"] == 500

    phase_config = project.create_config_for_phase("TurbineInstallation")

    assert phase_config["wtiv"]["name"] == "Example WTIV"
    assert phase_config["site"]["distance"] == 50

    project.run()


def test_expected_config_merging():
    """
    Tests for merging of expected configs
    """

    config1 = {
        "site": {"distance": "float", "depth": "float"},
        "plant": {"num_turbines": "int"},
    }

    config2 = {
        "site": {"distance": "float", "wave_height": "float"},
        "monopile": {"diameter": "float"},
    }

    config = ProjectManager.merge_dicts(config1, config2)

    assert config == {
        "site": {
            "distance": "float",
            "depth": "float",
            "wave_height": "float",
        },
        "plant": {"num_turbines": "int"},
        "monopile": {"diameter": "float"},
    }


def test_find_key_match():
    class SpecificTurbineInstallation(InstallPhase):
        expected_config = {}

    TestProjectManager = deepcopy(ProjectManager)
    TestProjectManager.register_install_phase(SpecificTurbineInstallation)

    phase_dict = TestProjectManager.phase_dict()
    assert "SpecificTurbineInstallation" in phase_dict.keys()

    tests = [
        ("TurbineInstallation", "TurbineInstallation"),
        ("TurbineInstallation_Test", "TurbineInstallation"),
        ("TurbineInstallation Test", "TurbineInstallation"),
        ("TurbineInstallation Test_1", "TurbineInstallation"),
        ("SpecificTurbineInstallation", "SpecificTurbineInstallation"),
        ("SpecificTurbineInstallation_Test", "SpecificTurbineInstallation"),
        ("SpecificTurbineInstallation Test", "SpecificTurbineInstallation"),
        ("SpecificTurbineInstallation Test_1", "SpecificTurbineInstallation"),
    ]

    for test in tests:
        i, expected = test
        response = TestProjectManager.find_key_match(i)

        assert response.__name__ == expected

    fails = [
        "DifferentTurbineInstallation",
        "Other TurbineInstallation",
        "Extra Different TurbineInstallation_1",
    ]

    for f in fails:
        assert TestProjectManager.find_key_match(f) is None


### Overlapping Install Phases
def test_install_phase_start_parsing():
    config_mixed_starts = deepcopy(config)
    config_mixed_starts["install_phases"] = {
        "MonopileInstallation": 0,
        "TurbineInstallation": "10/22/2009",
        "ArrayCableInstallation": ("MonopileInstallation", 0.5),
    }

    project = ProjectManager(config_mixed_starts, weather=weather_df)
    defined, depends = project._parse_install_phase_values(config_mixed_starts["install_phases"])
    assert len(defined) == 2
    assert len(depends) == 1

    assert defined["MonopileInstallation"] == 0
    assert defined["TurbineInstallation"] == 1


def test_chained_dependencies():
    config_chained = deepcopy(config)
    config_chained["spi_vessel"] = "test_scour_protection_vessel"
    config_chained["scour_protection"] = {
        "tonnes_per_substructure": 200,
        "cost_per_tonne": 45,
    }
    config_chained["install_phases"] = {
        "ScourProtectionInstallation": 0,
        "MonopileInstallation": ("ScourProtectionInstallation", 0.1),
        "TurbineInstallation": ("MonopileInstallation", 0.5),
    }

    project = ProjectManager(config_chained)
    project.run()

    df = pd.DataFrame(project.actions)
    sp = list(df.loc[df["phase"] == "ScourProtectionInstallation"]["time"])
    mp = list(df.loc[df["phase"] == "MonopileInstallation"]["time"])
    tu = list(df.loc[df["phase"] == "TurbineInstallation"]["time"])

    assert min(sp) == 0
    assert min(mp) == (max(sp) - min(sp)) * 0.1
    assert min(tu) == (max(mp) - min(mp)) * 0.5 + min(mp)


@pytest.mark.parametrize("m_start, t_start", [(0, 0), (0, 100), (100, 100), (100, 200)])
def test_index_starts(m_start, t_start):
    """
    Tests functionality related to passing index starts into 'install_phases' sub-dict.
    """
    _target_diff = t_start - m_start

    config_with_index_starts = deepcopy(config)
    config_with_index_starts["install_phases"] = {
        "MonopileInstallation": m_start,
        "TurbineInstallation": t_start,
    }

    project = ProjectManager(config_with_index_starts)
    project.run()

    df = pd.DataFrame(project.actions)

    _m = df.loc[df["phase"] == "MonopileInstallation"].iloc[0]
    _t = df.loc[df["phase"] == "TurbineInstallation"].iloc[0]

    _diff = (_t["time"] - _t["duration"]) - (_m["time"] - _m["duration"])
    assert _diff == _target_diff


@pytest.mark.parametrize(
    "m_start, t_start, expected",
    [
        (0, 0, 0),
        (0, 1000, 1000),
        (0, "05/01/2010", 4585),
        ("03/01/2010", "03/01/2010", 0),
        ("03/01/2010", "05/01/2010", 1464),
    ],
)
def test_start_dates_with_weather(m_start, t_start, expected):
    config_with_defined_starts = deepcopy(config)
    config_with_defined_starts["install_phases"] = {
        "MonopileInstallation": m_start,
        "TurbineInstallation": t_start,
    }

    project = ProjectManager(config_with_defined_starts, weather=weather_df)
    project.run()
    df = pd.DataFrame(project.actions)

    _m = df.loc[df["phase"] == "MonopileInstallation"].iloc[0]
    _t = df.loc[df["phase"] == "TurbineInstallation"].iloc[0]

    _diff = (_t["time"] - _t["duration"]) - (_m["time"] - _m["duration"])
    assert _diff == expected


def test_duplicate_phase_definitions():
    config_with_duplicates = deepcopy(config)
    config_with_duplicates["MonopileInstallation_1"] = {"plant": {"num_turbines": 5}}

    config_with_duplicates["MonopileInstallation_2"] = {
        "plant": {"num_turbines": 5},
        "site": {"distance": 100},
    }

    config_with_duplicates["install_phases"] = {
        "MonopileInstallation_1": 0,
        "MonopileInstallation_2": 800,
        "TurbineInstallation": 1600,
    }

    project = ProjectManager(config_with_duplicates)
    project.run()

    df = pd.DataFrame(project.actions).groupby(["phase", "action"]).count()["time"]

    assert df.loc[("MonopileInstallation_1", "Drive Monopile")] == 5
    assert df.loc[("MonopileInstallation_2", "Drive Monopile")] == 5
    assert df.loc[("TurbineInstallation", "Attach Tower Section")] == 10


def test_custom_install_phases():
    # Not a subclass
    class CustomInstallPhase:
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_design_phase(CustomInstallPhase)

    # Wrong subclass
    class CustomDesignPhase(DesignPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_install_phase(CustomDesignPhase)

    # Name already taken
    class MonopileInstallation(InstallPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_install_phase(MonopileInstallation)

    # Bad name format
    class MonopileInstallation_Custom(InstallPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_install_phase(MonopileInstallation_Custom)

    # Successful registration
    class CustomInstallPhase(InstallPhase):
        pass

    ProjectManager.register_install_phase(CustomInstallPhase)
    assert ProjectManager.find_key_match("CustomInstallPhase") == CustomInstallPhase


def test_custom_design_phases():
    # Not a subclass
    class CustomDesignPhase:
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_design_phase(CustomDesignPhase)

    # Wrong subclass
    class CustomInstallPhase(InstallPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_design_phase(CustomInstallPhase)

    # Name already taken
    class MonopileDesign(DesignPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_install_phase(MonopileDesign)

    # Bad name format
    class MonopileDesign_Custom(DesignPhase):
        pass

    with pytest.raises(ValueError):
        ProjectManager.register_install_phase(MonopileDesign_Custom)

    # Successful registration
    class CustomDesignPhase(DesignPhase):
        pass

    ProjectManager.register_design_phase(CustomDesignPhase)
    assert ProjectManager.find_key_match("CustomDesignPhase") == CustomDesignPhase


### Design Phase Interactions
def test_design_phases():
    config_with_design = deepcopy(config)

    # Add MonopileDesign
    config_with_design["design_phases"] = ["MonopileDesign"]

    # Add required parameters
    config_with_design["site"]["mean_windspeed"] = 9
    config_with_design["turbine"]["rotor_diameter"] = 200
    config_with_design["turbine"]["rated_windspeed"] = 10
    config_with_design["monopile_design"] = {}

    # Remove monopile sub dictionary
    _ = config_with_design.pop("monopile")
    project = ProjectManager(config_with_design)
    project.run()

    assert isinstance(project.config["monopile"], dict)

    project = ProjectManager(config_with_design)
    project.run()


### Outputs
def test_resolve_project_capacity():
    # Missing turbine rating
    config1 = {"plant": {"capacity": 600, "num_turbines": 40}}

    project1 = ProjectManager(config1)
    assert project1.config["plant"]["capacity"] == config1["plant"]["capacity"]
    assert project1.config["plant"]["num_turbines"] == config1["plant"]["num_turbines"]
    assert project1.config["turbine"]["turbine_rating"] == 15

    # Missing plant capacity
    config2 = {
        "plant": {"num_turbines": 40},
        "turbine": {"turbine_rating": 15},
    }

    project2 = ProjectManager(config2)
    assert project2.config["plant"]["capacity"] == 600
    assert project2.config["plant"]["num_turbines"] == config2["plant"]["num_turbines"]
    assert project2.config["turbine"]["turbine_rating"] == config2["turbine"]["turbine_rating"]

    # Missing number of turbines
    config3 = {"plant": {"capacity": 600}, "turbine": {"turbine_rating": 15}}

    project3 = ProjectManager(config3)
    assert project3.config["plant"]["capacity"] == config3["plant"]["capacity"]
    assert project3.config["plant"]["num_turbines"] == 40
    assert project3.config["turbine"]["turbine_rating"] == config3["turbine"]["turbine_rating"]

    # Test for float precision
    config4 = {
        "plant": {"capacity": 600, "num_turbines": 40},
        "turbine": {"turbine_rating": 15.0},
    }

    project4 = ProjectManager(config4)
    assert project4.config["plant"]["capacity"] == config4["plant"]["capacity"]
    assert project4.config["plant"]["num_turbines"] == config4["plant"]["num_turbines"]
    assert project4.config["turbine"]["turbine_rating"] == config4["turbine"]["turbine_rating"]

    # Non matching calculated value
    config5 = {
        "plant": {"capacity": 700, "num_turbines": 40},
        "turbine": {"turbine_rating": 15.0},
    }

    with pytest.raises(AttributeError):
        _ = ProjectManager(config5)

    # Test for not enough information
    config6 = {"plant": {"capacity": 600}}

    project6 = ProjectManager(config6)
    assert project6.config["plant"]["capacity"] == config6["plant"]["capacity"]

    with pytest.raises(KeyError):
        _ = project6.config["turbine"]["turbine_rating"]

    with pytest.raises(KeyError):
        _ = project6.config["plant"]["num_turbines"]


### Exceptions
def test_incomplete_config():
    incomplete_config = deepcopy(config)
    _ = incomplete_config["site"].pop("depth")

    with pytest.raises(MissingInputs):
        project = ProjectManager(incomplete_config)
        project.run()


def test_wrong_phases():
    wrong_phases = deepcopy(config)
    wrong_phases["install_phases"].append("IncorrectPhaseName")

    with pytest.raises(PhaseNotFound):
        project = ProjectManager(wrong_phases)
        project.run()


def test_bad_dates():
    bad_dates = deepcopy(config)
    bad_dates["install_phases"] = {
        "MonopileInstallation": "03/01/2015",
        "TurbineInstallation": "05/01/2015",
    }

    with pytest.raises(WeatherProfileError):
        project = ProjectManager(bad_dates, weather=weather_df)
        project.run()


def test_no_defined_start():
    missing_start = deepcopy(config)
    missing_start["install_phases"] = {
        "MonopileInstallation": ("TurbineInstallation", 0.1),
        "TurbineInstallation": ("MonopileInstallation", 0.1),
    }

    with pytest.raises(ValueError):
        project = ProjectManager(missing_start)
        project.run()


def test_circular_dependencies():
    circular_deps = deepcopy(config)
    circular_deps["spi_vessel"] = "test_scour_protection_vessel"
    circular_deps["scour_protection"] = {
        "tonnes_per_substructure": 200,
        "cost_per_tonne": 45,
    }
    circular_deps["install_phases"] = {
        "ScourProtectionInstallation": 0,
        "MonopileInstallation": ("TurbineInstallation", 0.1),
        "TurbineInstallation": ("MonopileInstallation", 0.1),
    }

    with pytest.raises(PhaseDependenciesInvalid):
        project = ProjectManager(circular_deps)
        project.run()


def test_dependent_phase_ordering():
    wrong_order = deepcopy(config)
    wrong_order["spi_vessel"] = "test_scour_protection_vessel"
    wrong_order["scour_protection"] = {
        "tonnes_per_substructure": 200,
        "cost_per_tonne": 45,
    }
    wrong_order["install_phases"] = {
        "ScourProtectionInstallation": ("TurbineInstallation", 0.1),
        "TurbineInstallation": ("MonopileInstallation", 0.1),
        "MonopileInstallation": 0,
    }

    project = ProjectManager(wrong_order)
    project.run()

    assert len(project.phase_times) == 3


def test_ProjectProgress():
    data = [
        ("Export System", 10),
        ("Offshore Substation", 20),
        ("Array String", 15),
        ("Array String", 25),
        ("Turbine", 5),
        ("Turbine", 10),
        ("Turbine", 15),
        ("Turbine", 20),
        ("Turbine", 25),
        ("Substructure", 6),
        ("Substructure", 9),
        ("Substructure", 14),
        ("Substructure", 22),
        ("Substructure", 26),
    ]

    progress = ProjectProgress(data)

    assert progress.parse_logs("Export System") == [10]

    turbines = progress.parse_logs("Turbine")
    assert len(turbines) == 5

    chunks = list(progress.chunk_max(turbines, 2))
    assert chunks[0] == 10
    assert chunks[1] == 20
    assert chunks[2] == 25

    assert progress.complete_export_system == 20
    times, _ = progress.complete_array_strings
    assert times == [15, 26]

    times, turbines = progress.energize_points
    assert times == [20, 26]
    assert sum(turbines) == 5


def test_ProjectProgress_with_incomplete_project():
    project = ProjectManager(config)
    project.run()

    _ = project.progress.parse_logs("Substructure")
    _ = project.progress.parse_logs("Turbine")

    with pytest.raises(ValueError):
        project.progress.complete_export_system

    with pytest.raises(ValueError):
        project.progress.complete_array_strings


def test_ProjectProgress_with_complete_project():
    project = ProjectManager(complete_project)
    project.run()

    _ = project.progress.parse_logs("Substructure")
    _ = project.progress.parse_logs("Turbine")
    _ = project.progress.parse_logs("Array String")
    _ = project.progress.parse_logs("Export System")
    _ = project.progress.parse_logs("Offshore Substation")

    _ = project.progress.complete_export_system
    _ = project.progress.complete_array_strings

    _ = project.progress.energize_points

    new = deepcopy(complete_project)
    new["plant"]["num_turbines"] = 61

    # Uneven strings
    project = ProjectManager(new)
    project.run()

    _ = project.progress.energize_points


def test_monthly_expenses():
    project = ProjectManager(complete_project)
    project.run()
    _ = project.monthly_expenses

    # Still report expenses for "incomplete" project
    config = deepcopy(complete_project)
    _ = config["install_phases"].pop("TurbineInstallation")

    project = ProjectManager(config)
    project.run()

    _ = project.monthly_expenses


def test_monthly_revenue():
    project = ProjectManager(complete_project)
    project.run()
    _ = project.monthly_revenue

    # Can't generate revenue with "incomplete" project
    config = deepcopy(complete_project)
    _ = config["install_phases"].pop("TurbineInstallation")

    project = ProjectManager(config)
    project.run()

    with pytest.raises(ValueError):
        _ = project.monthly_revenue


def test_cash_flow():
    project = ProjectManager(complete_project)
    project.run()
    _ = project.cash_flow

    # Can't generate revenue with "incomplete" project but cash flow will still
    # be reported
    config = deepcopy(complete_project)
    _ = config["install_phases"].pop("TurbineInstallation")

    project = ProjectManager(config)
    project.run()

    cash_flow = project.cash_flow
    assert all(v <= 0 for v in cash_flow.values())


def test_npv():
    project = ProjectManager(complete_project)
    project.run()
    baseline = project.npv

    config = deepcopy(complete_project)
    config["project_parameters"] = {"ncf": 0.35}
    project = ProjectManager(config)
    project.run()
    assert project.npv != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"offtake_price": 70}
    project = ProjectManager(config)
    project.run()
    assert project.npv != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"project_lifetime": 30}
    project = ProjectManager(config)
    project.run()
    assert project.npv != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"discount_rate": 0.03}
    project = ProjectManager(config)
    project.run()
    assert project.npv != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"opex_rate": 120}
    project = ProjectManager(config)
    project.run()
    assert project.npv != baseline


def test_soft_costs():
    project = ProjectManager(complete_project)
    baseline = project.soft_capex

    config = deepcopy(complete_project)
    config["project_parameters"] = {"construction_insurance": 50}
    project = ProjectManager(config)
    assert project.soft_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"construction_financing": 190}
    project = ProjectManager(config)
    assert project.soft_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"contingency": 320}
    project = ProjectManager(config)
    assert project.soft_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"contingency": 320}
    project = ProjectManager(config)
    assert project.soft_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"commissioning": 50}
    project = ProjectManager(config)
    assert project.soft_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"decommissioning": 50}
    project = ProjectManager(config)
    assert project.soft_capex != baseline


def test_project_costs():
    project = ProjectManager(complete_project)
    baseline = project.project_capex

    config = deepcopy(complete_project)
    config["project_parameters"] = {"site_auction_price": 50e6}
    project = ProjectManager(config)
    assert project.project_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"site_assessment_cost": 25e6}
    project = ProjectManager(config)
    assert project.project_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"construction_plan_cost": 25e6}
    project = ProjectManager(config)
    assert project.project_capex != baseline

    config = deepcopy(complete_project)
    config["project_parameters"] = {"installation_plan_cost": 25e6}
    project = ProjectManager(config)
    assert project.project_capex != baseline


def test_capex_categories():
    project = ProjectManager(complete_project)
    project.run()
    baseline = project.capex_breakdown
    _ = project.capex_breakdown_per_kw

    new_config = deepcopy(complete_project)
    new_config["install_phases"]["ExportCableInstallation_1"] = 0
    new_project = ProjectManager(new_config)
    new_project.run()
    new_breakdown = new_project.capex_breakdown

    assert new_breakdown["Export System"] > baseline["Export System"]
    assert new_breakdown["Export System Installation"] > baseline["Export System Installation"]



import os
import sys
import copy
import platform
import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from wisdem.optimization_drivers.intermittent_component import IntermittentComponent

rosenbrock_size = 4
num_procs = 1


class Rosenbrock1(IntermittentComponent):
    def initialize(self):
        # This is required only if you have something to add to the initialize(),
        # such as OpenMDAO options declarations
        super().initialize()
        self.options.declare("multiplier", 1.0)

    def setup(self):
        # This is required to run the original setup() contained within IntermittentComponent
        super().setup()

        self.add_input("x", np.ones(rosenbrock_size))
        self.add_output("f1", 0.0)

    def internal_compute(self, inputs, outputs):
        """
        This is the actual method where the computations should occur.
        """
        x = inputs["x"]
        x_0 = x[:-1]
        outputs["f1"] = self.options["multiplier"] * sum((1 - x_0) ** 2)


class Rosenbrock2(IntermittentComponent):
    def initialize(self):
        # This is required only if you have something to add to the initialize(),
        # such as OpenMDAO options declarations
        super().initialize()
        self.options.declare("multiplier", 100.0)

    def setup(self):
        # This is required to run the original setup() contained within IntermittentComponent
        super().setup()

        self.add_input("x", np.ones(rosenbrock_size))
        self.add_output("f2", 0.0)

    def internal_compute(self, inputs, outputs):
        """
        This is the actual method where the computations should occur.
        """
        x = inputs["x"]
        x_0 = x[:-1]
        x_1 = x[1:]
        outputs["f2"] = self.options["multiplier"] * sum((x_1 - x_0**2) ** 2)


class TestIntermittentComponent(unittest.TestCase):
    def test_run(self):
        prob = om.Problem(reports=False, model=om.Group(num_par_fd=num_procs))
        prob.model.approx_totals(method="fd")
        indeps = prob.model.add_subsystem("indeps", om.IndepVarComp(), promotes=["*"])
        indeps.add_output("x", 1.2 * np.ones(rosenbrock_size))

        prob.model.add_subsystem(
            "rosenbrock1",
            Rosenbrock1(num_iterations_between_calls=10, multiplier=1.0),
            promotes=["*"],
        )
        prob.model.add_subsystem(
            "rosenbrock2",
            Rosenbrock2(num_iterations_between_calls=1, multiplier=100.0),
            promotes=["*"],
        )
        prob.model.add_subsystem("objective_comp", om.ExecComp("f = f1 + f2"), promotes=["*"])

        # setup the optimization
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options["optimizer"] = "SLSQP"
        prob.driver.opt_settings["tol"] = 1e-9

        prob.model.add_design_var("x", lower=-1.5, upper=1.5)
        prob.model.add_objective("f")

        prob.setup()
        prob.run_driver()

        assert_near_equal(prob["f"][0], 0.0, tolerance=1e-2)
        assert_near_equal(prob["x"], np.ones(4), tolerance=1e-1)
        self.assertEqual(prob.model.rosenbrock1.actual_compute_calls, 6)
        # Seems sensivity to backend SLSQP parameter tuning
        # self.assertEqual(prob.model.rosenbrock2.actual_compute_calls, 48)

        # minimum value
        print(f"Optimum found: {prob['f'][0]}")
        # location of the minimum
        print(f"Optimal design: {prob['x']}")
        print()
        print(
            f"Number of actual compute calls: {prob.model.rosenbrock1.actual_compute_calls} and {prob.model.rosenbrock2.actual_compute_calls}"
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8

import numpy as np

import wisdem.pyframe3dd.pyframe3dd as pyframe3dd

# 0 ------- node data ----------------
inode = np.array([1, 2, 3, 4, 5])
x = np.array([0.0, -1200, 1200, 1200, -1200])
y = np.array([0.0, -900, -900, 900, 900])
z = np.array([1000.0, 0.0, 0.0, 0.0, 0.0])
r = np.array([0.0, 0.0, 0.0, 0.0, 0.0])

nodes = pyframe3dd.NodeData(inode, x, y, z, r)
# 0 -----------------------------------

# 1 ------ reaction data ------------
rnode = np.array([2, 3, 4, 5])
Rx = np.ones(4)
Ry = np.ones(4)
Rz = np.ones(4)
Rxx = np.ones(4)
Ryy = np.ones(4)
Rzz = np.ones(4)

reactions = pyframe3dd.ReactionData(rnode, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=1)
# 1 -----------------------------------

# 2 ------ frame element data ------------
ielement = np.array([1, 2, 3, 4])
N1 = np.array([2, 1, 1, 5])
N2 = np.array([1, 3, 4, 1])
Ax = 36.0 * np.ones(4)
Asy = 20.0 * np.ones(4)
Asz = 20.0 * np.ones(4)
Jx = 1000.0 * np.ones(4)
Iy = 492.0 * np.ones(4)
Iz = 492.0 * np.ones(4)
E = 200000.0 * np.ones(4)
G = 79300.0 * np.ones(4)
roll = np.zeros(4)
density = 7.85e-9 * np.ones(4)


elements = pyframe3dd.ElementData(ielement, N1, N2, Ax, Asy, Asz, Jx, Iy, Iz, E, G, roll, density)
# 2 -----------------------------------


# 3 ------ other data ------------
shear = True  # 1: include shear deformation
geom = True  # 1: include geometric stiffness
dx = 20.0  # x-axis increment for internal forces
other = pyframe3dd.Options(shear, geom, dx)
# 3 -----------------------------------


# 4 ------ initialize frame3dd object
frame = pyframe3dd.Frame(nodes, reactions, elements, other)
# 4 -----------------------------------


# 5 ------ static load case 1 ------------
# gravity in the X, Y, Z, directions (global)
gx = 0.0
gy = 0.0
gz = -9806.33
load = pyframe3dd.StaticLoadCase(gx, gy, gz)

# point load
nF = np.array([1])
Fx = np.array([100.0])
Fy = np.array([-200.0])
Fz = np.array([-100.0])
Mxx = np.array([0.0])
Myy = np.array([0.0])
Mzz = np.array([0.0])
load.changePointLoads(nF, Fx, Fy, Fz, Mxx, Myy, Mzz)

frame.addLoadCase(load)
# 5 -----------------------------------


# 6 ------ static load case 2 ------------
# gravity in the X, Y, Z, directions (global)
gx = 0.0
gy = 0.0
gz = -9806.33
load = pyframe3dd.StaticLoadCase(gx, gy, gz)

# uniform loads
EL = np.array([2, 1])
Ux = np.array([0.0, 0.0])
Uy = np.array([0.1, 0.0])
Uz = np.array([0.0, 0.1])
load.changeUniformLoads(EL, Ux, Uy, Uz)

# trapezoidally distributed loads
EL = np.array([3, 4])
xx1 = np.array([20.0, 0.0])
xx2 = np.array([80.0, 0.0])
wx1 = np.array([0.01, 0.0])
wx2 = np.array([0.05, 0.0])
xy1 = np.array([0.0, 68.0])
xy2 = np.array([0.0, 330.0])
wy1 = np.array([0.0, 0.05])
wy2 = np.array([0.0, 0.0])
xz1 = np.array([80.0, 80.0])
xz2 = np.array([830.0, 830.0])
wz1 = np.array([-0.05, -0.05])
wz2 = np.array([0.07, 0.07])
load.changeTrapezoidalLoads(EL, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

EL = np.array([1])
a = np.array([12e-6])
hy = np.array([10.0])
hz = np.array([10.0])
Typ = np.array([20.0])
Tym = np.array([10.0])
Tzp = np.array([10.0])
Tzm = np.array([-10.0])
load.changeTemperatureLoads(EL, a, hy, hz, Typ, Tym, Tzp, Tzm)

frame.addLoadCase(load)
# 6 -----------------------------------


# 7 ------ static load case 3 ------------
gx = 0.0
gy = 0.0
gz = -9806.33
load = pyframe3dd.StaticLoadCase(gx, gy, gz)

# concentrated interior point loads
EL = np.array([1, 2])
Px = np.array([0.0, 0.0])
Py = np.array([100.0, -200.0])
Pz = np.array([-900.0, 200.0])
x = np.array([600.0, 800.0])
load.changeElementLoads(EL, Px, Py, Pz, x)

frame.addLoadCase(load)
# 7 -----------------------------------


# 8 ------ dyamic analysis data ------------
nM = 6  # number of desired dynamic modes of vibration
Mmethod = 1  # 1: subspace Jacobi     2: Stodola
lump = 0  # 0: consistent mass ... 1: lumped mass matrix
tol = 1e-9  # mode shape tolerance
shift = 0.0  # shift value ... for unrestrained structures
frame.enableDynamics(nM, Mmethod, lump, tol, shift)
# 8 -----------------------------------

# 9 ---- extra node inertia data
N = np.array([1])
EMs = np.array([0.1])
EMx = np.array([0.0])
EMy = np.array([0.0])
EMz = np.array([0.0])

# frame.changeExtraInertia(N, EMs, EMx, EMy, EMz)
frame.changeExtraNodeMass(N, EMs, EMx, EMy, EMz, [0.0], [0.0], [0.0], [0.0], [0.0], [0.0], False)
# 9 ------------------------------------

# 10 ----- run the analysis
displacements, forces, reactions, internalForces, mass, modal = frame.run()
# 10 ------------------------------------

# 11 ----- analysis output
nC = len(frame.loadCases)  # number of load cases
nN = len(nodes.node)  # number of nodes
nE = len(elements.element)  # number of elements

# mass data
print("total_mass =", mass.total_mass)
print("struct_mass =", mass.struct_mass)
print("node =", mass.node)
print("xmass =", mass.xmass)
print("ymass =", mass.ymass)
print("zmass =", mass.zmass)
print("xinrta =", mass.xinrta)
print("yinrta =", mass.yinrta)
print("zinrta =", mass.zinrta)
print()
print()

# node displacements
for iCase in range(nC):
    print("case_idx:", iCase)

    print("node:", displacements.node[iCase, :])
    print("dx:", displacements.dx[iCase, :])
    print("dy:", displacements.dy[iCase, :])
    print("dz:", displacements.dz[iCase, :])
    print("dxrot:", displacements.dxrot[iCase, :])
    print("dyrot:", displacements.dyrot[iCase, :])
    print("dzrot:", displacements.dzrot[iCase, :])
    print()
    print("element =", forces.element[iCase, :])
    print("node =", forces.node[iCase, :])
    print("Nx =", forces.Nx[iCase, :])
    print("Vy =", forces.Vy[iCase, :])
    print("Vz =", forces.Vz[iCase, :])
    print("Txx =", forces.Txx[iCase, :])
    print("Myy =", forces.Myy[iCase, :])
    print("Mzz =", forces.Mzz[iCase, :])
    print()
    print("nodesR =", reactions.node[iCase, :])
    print("RFx =", reactions.Fx[iCase, :])
    print("RFy =", reactions.Fy[iCase, :])
    print("RFz =", reactions.Fz[iCase, :])
    print("RMxx =", reactions.Mxx[iCase, :])
    print("RMyy =", reactions.Myy[iCase, :])
    print("RMzz =", reactions.Mzz[iCase, :])
    print()

print()
print()

# internal forces
iE = 3  # note just showing for one element
for iCase in range(nC):
    print("case_idx:", iCase)
    print("element_idx:", iE)

    print("x =", internalForces[iE].x[iCase, :])
    print("Nx =", internalForces[iE].Nx[iCase, :])
    print("Vy =", internalForces[iE].Vy[iCase, :])
    print("Vz =", internalForces[iE].Vz[iCase, :])
    print("Tx =", internalForces[iE].Tx[iCase, :])
    print("My =", internalForces[iE].My[iCase, :])
    print("Mz =", internalForces[iE].Mz[iCase, :])
    print("Dx =", internalForces[iE].Dx[iCase, :])
    print("Dy =", internalForces[iE].Dy[iCase, :])
    print("Dz =", internalForces[iE].Dz[iCase, :])
    print("Rx =", internalForces[iE].Rx[iCase, :])
    print()

print()
print()

# Modal data
for iM in range(nM):
    print("mode_idx", iM)

    print("freq =", modal.freq[iM])
    print("xmpf =", modal.xmpf[iM])
    print("ympf =", modal.ympf[iM])
    print("zmpf =", modal.zmpf[iM])
    print("node =", modal.node[iM, :])
    print("xdsp =", modal.xdsp[iM, :])
    print("ydsp =", modal.ydsp[iM, :])
    print("zdsp =", modal.zdsp[iM, :])
    print("xrot =", modal.xrot[iM, :])
    print("yrot =", modal.yrot[iM, :])
    print("zrot =", modal.zrot[iM, :])
    print()
# 11 ------------------------------------

# 12 ------------------------------------
# frame.write("pyramid_frame_B.3dd")
# 12 ------------------------------------
